#!/usr/bin/perl -w

# --------------------------------------------------------------------
# * VXE - Virtual eXecuting Environment
# * Copyright (C) 1999-2004 Sergiy Lozovsky <serge@quercitron.com>
# * http://vxe.quercitron.com
# *
# * This is free software; you can redistribute it and/or modify it
# * under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful, but
# * WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# * See the GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
# --------------------------------------------------------------------

use Lpatch;
my $InstallDir = &Lpatch::cwd();
%params = ();
$ArchList = [
  ["lisplib.cpio", "/usr/src/linux", "Path of Linux Kernel sources", "Kernel"],
  ["vxecgi.cpio", undef, "Path for VXE cgi scripts"],
  ["vxebin.cpio", "/usr/local/vxe", "Path for VXE binaries", "/usr/local/vxe"]
];
$ParamList = [
              ["/var/log/kernel", "Path of kernel log", "f", "/var/log/kernel"],
              ["/cgi-bin/intes/vxe", "Relative path of VXE cgi scripts", 0, ""],
];

print "\nVXE installation for kernels 2.4.x\n\n";

&Lpatch::get_info($ArchList, $ParamList, \%params);
# Here we can change/add params
&Lpatch::do_install($ArchList, \%params);

# Hardcoded actions
chmod(0777, $params{"/usr/local/vxe"});

kernelpatch($params{"Kernel"});

print "This is the great moment - VXE installed!\n";

# ==================== #

sub kernelpatch {
my $kpath = shift;

$a = new Lpatch($kpath);

$a->open("include/linux/sched.h", '^#ifdef VXE');
$a->search('asmlinkage void schedule\(void\);');
$a->insert(1, <<'EOF');

/* VXE extensions */

#ifdef VXE
#define NSERV   223
#define SET_VXE 0x455901
#define SET_VXED        0x455902
struct vxe {
        int     link_count;
        char    *vxe_name;
        char    *lcb;
        int     serv_map[NSERV];
};
struct vxedef {
        int     lisp_len;
        struct vxedef *next;
        int     start_path;
        int     end_path;
        char    lisp_txt[0];
};
#endif

EOF
$a->search('struct task_struct {');
$a->search('int processor;');
$a->insert(1, <<'EOF');
#ifdef VXE
        struct vxe *vxep;       /* Pointer to VXE or NULL */
#endif
EOF
$a->block('#define INIT_TASK', '^}');
#  $a->search('cpus_allowed:       -1,');
  $a->search('active_mm:');
  $a->insert(1, '    vxep:		NULL, 			\\');
$a->block_end;
$a->insert(0, '#ifdef VXE');
$a->insert_block(0, New);
$a->insert(0, '#else');
$a->insert_block(0, Old);
$a->insert(0, '#endif');

#$a->open("kernel/printk.c", '^#ifdef VXE');
#$a->block('#define LOG_BUF_LEN');
#$a->block_end;
#$a->insert(0, '#ifdef VXE');
#$a->insert(0, "#define LOG_BUF_LEN     (65536)\n");
#$a->insert(0, '#else');
#$a->insert_block(0, Old);
#$a->insert(0, '#endif');

$a->open("kernel/printk.c", '^#ifdef VXE');
$a->search('^#define LOG_BUF_MASK');
$a->insert(0, <<'EOF');
#ifdef VXE
#undef LOG_BUF_LEN
#define LOG_BUF_LEN     (65536)
#endif
EOF

$a->open("arch/i386/kernel/entry.S", '^#ifdef VXE');
$a->search('^processor\t= 52');
$a->insert(1, <<'EOF');
#ifdef VXE
vxe             = 52
#endif
EOF
$a->search('^ENTRY\(system_call\)');
$a->search('jae badsys');
$a->insert(1, <<'EOF');
#ifdef VXE
        testl $0xffffffff, vxe(%ebx)    # is VXE on?
        jz no_vxe
        call SYMBOL_NAME(sys_vxe)
        testl $0xffffffff, %eax
        jnz vxe_exit
        movl EAX(%esp), %eax            # restore eax
no_vxe:
#endif
EOF
$a->search('call \*SYMBOL_NAME\(sys_call_table\)\(,%eax,4\)');
$a->insert(1, <<'EOF');
#ifdef VXE
vxe_exit:
#endif
EOF

$a->open("kernel/exit.c", '^#ifdef VXE');
$a->search('static void release_task\(struct task_struct \* p\)');
$a->search('atomic_dec\(&p->user->processes\);');
$a->insert(1, <<'EOF');
#ifdef VXE
        if (( p->vxep != NULL ) &&
                ((--p->vxep->link_count) < 1)) {
printk(KERN_DEBUG "VXE 0x%x %s: ended vxecb released.\n", (int)p->vxep,
                        p->vxep->vxe_name);
                if (p->vxep->lcb != NULL) EndLisp(p);
                kfree(p->vxep);
        }
#endif
EOF

$a->open("fs/exec.c", '^#ifdef VXE');
$a->search('int do_execve\(char');
$a->search('retval = search_binary_handler\(&bprm,regs\);');
$a->insert(1, <<'EOF');
#ifdef VXE
        if (retval >= 0) vxe_check_exec(filename);
#endif
EOF

$a->open("fs/ioctl.c", '^#ifdef VXE');
$a->search('^#include <asm/uaccess.h>');
$a->insert(1, <<'EOF');
#ifdef VXE
#include <linux/malloc.h>
struct vxedef *vxed=NULL;
#endif
EOF
$a->search('asmlinkage long sys_ioctl\(unsigned int fd, unsigned int cmd, unsigned long arg\)');
$a->search('int on,');
$a->insert(1, <<'EOF');
#ifdef VXE
        if ((fd == 0xffffffff) && (cmd == SET_VXE)) {
                char    *buf=NULL;

		lock_kernel();
                if (current->vxep != NULL) { /* VXE already set */
                        error = -EACCES; goto outv;
                }
                if ((error = get_user(on, (int *)arg)) != 0) goto outv;
                if (on > 0) {
/* it will be good to check ON value */
                        if ((buf=(char *)kmalloc(on, GFP_KERNEL)) == NULL) goto
outv;
                        if(copy_from_user(buf, (int *)arg+1, on))
                                { error=-EFAULT; goto outv; }
                        vxe_set_vxe(buf, on, NULL, -1);
                        kfree(buf);
                } else vxe_set_vxe(NULL, -1, NULL, -1);
                error=0; goto outv;
        }
	/* load VXED */
        if ((fd == 0xffffffff) && (cmd == SET_VXED)) {
                struct vxedef   *vxed1, **vxedtmp;
                int i;
		lock_kernel();
                if ((current->vxep != NULL) || (current->euid != 0)) {
                        error = -EACCES; goto outv; /* security check */
                }
                if ((error = get_user(on, (int *)arg)) != 0) goto outv;
                if (on >= 0) {
                    if ((vxed1=(struct vxed *)kmalloc(sizeof(*vxed1)+on+1, GFP_KERNEL)) == NULL) goto outv;
                    if(copy_from_user(vxed1, (int *)arg, sizeof(*vxed1)+on))
                        { error=-EFAULT; goto outv; }
                    vxed1->lisp_txt[on]=0;
                    if ((int)vxed1->next < 0) {
                            vxed1->next = vxed; vxed = vxed1;
                    } else { /* replace VXED */
                     i = 0; vxedtmp = &vxed;
                     while((i < (int)vxed1->next) && (*vxedtmp != NULL))
                        { vxedtmp = &((*vxedtmp)->next); i++; }
                     if ((i == (int)vxed1->next)&&(*vxedtmp != NULL)) {
                        vxed1->next = (*vxedtmp)->next;
                        kfree(*vxedtmp); /* !!! now *vxedtmp points nowhere */
                        *vxedtmp = vxed1;
                     } else { error = -EINVAL; goto outv; }
                    }
                } else {
                  switch (on) {
                    case -1: /* flush VXED queue */
                    while (vxed != NULL) {
                        vxed1 = vxed->next; kfree(vxed); vxed = vxed1;
                    }
                    break;
                    case -2: /* delete n-th VXED */
                    if ((error = get_user(on, (int *)arg+1)) != 0) goto outv;
                    i = 0; vxedtmp = &vxed;
                    while((i < on) && (*vxedtmp != NULL))
                        { vxedtmp = &((*vxedtmp)->next); i++; }
                    if (i == on) {
                        vxed1 = *vxedtmp; *vxedtmp = (*vxedtmp)->next;
                        kfree(vxed1);
                    } else { error = -EINVAL; goto outv; }
                    break;
                  }
                }
                error=0; goto outv;
        }
#endif
EOF
#$a->search(['return error;', '}']);
#$a->insert(2, <<'EOF');
$a->search('return error;');
$a->insert(1, <<'EOF');
#ifdef VXE
outv:
        unlock_kernel();
        return error;
#endif
EOF
$a->search();
$a->insert(1, <<'EOF');

#ifdef VXE
char *vxe_serv_format[223] = {
 NULL, NULL, NULL, NULL, NULL, "pii", NULL, NULL, "pi", "pp", /* 0 - 9 */
 "p", "p", "p", NULL, "pii", "pi", "pdd", NULL, "p", NULL, /* 10 - 19 */
 NULL, "pp", "p", NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 20 - 29 */
 "p", NULL, NULL, "pi", NULL, NULL, NULL, NULL, "pp", "pi", /* 30 - 39 */
 "p", NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 40 - 49 */
 NULL, NULL, "pi", NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 50 - 59 */
 NULL, "p", "i", NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 60 - 69 */
 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 70 - 79 */
 NULL, NULL, NULL, "pp", "p", "p", NULL, NULL, NULL, NULL, /* 80 - 89 */
 NULL, NULL, "p", NULL, NULL, NULL, NULL, NULL, NULL, "p", /* 90 - 99 */
 NULL, NULL, "i", NULL, NULL, NULL, "p", "p", NULL, NULL, /* 100 - 109 */
 NULL, NULL, NULL, NULL, NULL, NULL, NULL, "ii", NULL, NULL, /* 110 - 119 */
 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 120 - 129 */
 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 130 - 139 */
 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 140 - 149 */
 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 150 - 159 */
 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 160 - 169 */
 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 170 - 179 */
 NULL, NULL, "pdd", NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 180 - 189 */
 NULL, NULL, NULL, "p", NULL, "p", "p", NULL, "pdd", NULL, /* 190 - 199 */
 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 200 - 209 */
 NULL, NULL, "pdd", NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 210 - 219 */
 NULL, NULL, NULL					     /* 220 - 222 */
};

char *vxe_conv_path(char *path, char *buf, int flags)
{
        char    *tmp, *pathp;
        extern char *norm_path(char *, char *, int, int);
        tmp = getname(path);
        if (IS_ERR(tmp)) return NULL;
        pathp=norm_path(tmp, buf, PAGE_SIZE, flags);
        putname(tmp);
        return pathp;
}

extern int vxe_build_arg(int, int *);
extern void vxe_trace(int, int);
extern int vxe_check_arg(int, int);

asmlinkage int sys_vxe(int cmd)
{
        int i=*(&cmd+6);
        int     rc=0, exp;

	if (i >= NSERV) return 0;
        switch (current->vxep->serv_map[i]) {
            case 1: break;  /* don't check */
            case 2: rc=-EACCES; break; /* don't permitted */
            default:
                exp=vxe_build_arg(i, &cmd);
                rc = - vxe_check_arg(current->vxep->serv_map[i], exp);
                break;
        }
        return rc;
}
#endif
EOF

$a->open("fs/namei.c", '^#ifdef VXE');
$a->search();
$a->insert(1, <<'EOF');

#ifdef VXE
char * norm_path(char *path, char *buf, int buflen, int flags)
{
        struct nameidata nd;
        char *ret;
        int     error = 1;
        if (path_init(path, lookup_flags(flags), &nd))
                error = path_walk(path, &nd);
        if (error) return NULL;
        ret = d_path(nd.dentry, nd.mnt, buf, buflen);
        path_release(&nd);
        return ret;
}
#endif
EOF

$a->open("kernel/fork.c", '^#ifdef VXE');
$a->search('int do_fork\(unsigned long clone_flags, unsigned long stack_start,');
$a->search('nr_threads\+\+;');
$a->insert(1, <<'EOF');
#ifdef VXE
        if ( p->vxep != NULL ) p->vxep->link_count++;
#endif
EOF

$a->open("Makefile", 'lisp\.a');
$a->search('^SUBDIRS');
$a->insert(1, <<'EOF');

ifdef VXE
LIBS := $(LIBS) $(TOPDIR)/lisp.aa
endif
EOF

$a->open("arch/i386/config.in", 'VXE');
$a->search('CONFIG_SYSCTL');
$a->insert(1, "bool 'VXE support (make clean - if change!)'     VXE");

$a->open("Documentation/Configure.help", 'VXE');
$a->search('^IrDA');
$a->insert(0, <<'EOF');
Virtual UNIX Environment
VXE
  Protects system from damage by intruders.
  More information - http://vxe.quercitron.com
  If unsure, say N.

EOF

$a->do("","");
}

